package java.util.concurrent.atomic;

import java.util.function.LongBinaryOperator;
import java.util.function.LongUnaryOperator;
import sun.misc.Unsafe;
import java.io.Serializable;

public class AtomicLong extends Number implements Serializable
{
    private static final long serialVersionUID = 1927816293512124184L;
    private static final Unsafe unsafe;
    private static final long valueOffset;
    static final boolean VM_SUPPORTS_LONG_CAS;
    private volatile long value;
    
    private static native boolean VMSupportsCS8();
    
    public AtomicLong(long value) {
        this.value = value;
    }
    
    public AtomicLong() {
    }
    
    public final long get() {
        return this.value;
    }
    
    public final void set(long value) {
        this.value = value;
    }
    
    public final void lazySet(long x) {
        AtomicLong.unsafe.putOrderedLong(this, AtomicLong.valueOffset, x);
    }
    
    public final long getAndSet(long newValue) {
        return AtomicLong.unsafe.getAndSetLong(this, AtomicLong.valueOffset, newValue);
    }
    
    public final boolean compareAndSet(long expected, long x) {
        return AtomicLong.unsafe.compareAndSwapLong(this, AtomicLong.valueOffset, expected, x);
    }
    
    public final boolean weakCompareAndSet(long expected, long x) {
        return AtomicLong.unsafe.compareAndSwapLong(this, AtomicLong.valueOffset, expected, x);
    }
    
    public final long getAndIncrement() {
        return AtomicLong.unsafe.getAndAddLong(this, AtomicLong.valueOffset, 1L);
    }
    
    public final long getAndDecrement() {
        return AtomicLong.unsafe.getAndAddLong(this, AtomicLong.valueOffset, -1L);
    }
    
    public final long getAndAdd(long delta) {
        return AtomicLong.unsafe.getAndAddLong(this, AtomicLong.valueOffset, delta);
    }
    
    public final long incrementAndGet() {
        return AtomicLong.unsafe.getAndAddLong(this, AtomicLong.valueOffset, 1L) + 1L;
    }
    
    public final long decrementAndGet() {
        return AtomicLong.unsafe.getAndAddLong(this, AtomicLong.valueOffset, -1L) - 1L;
    }
    
    public final long addAndGet(long delta) {
        return AtomicLong.unsafe.getAndAddLong(this, AtomicLong.valueOffset, delta) + delta;
    }
    
    public final long getAndUpdate(LongUnaryOperator longUnaryOperator) {
        long value;
        do {
            value = this.get();
        } while (!this.compareAndSet(value, longUnaryOperator.applyAsLong(value)));
        return value;
    }
    
    public final long updateAndGet(LongUnaryOperator longUnaryOperator) {
        long value;
        long applyAsLong;
        do {
            value = this.get();
            applyAsLong = longUnaryOperator.applyAsLong(value);
        } while (!this.compareAndSet(value, applyAsLong));
        return applyAsLong;
    }
    
    public final long getAndAccumulate(long n, LongBinaryOperator longBinaryOperator) {
        long value;
        do {
            value = this.get();
        } while (!this.compareAndSet(value, longBinaryOperator.applyAsLong(value, n)));
        return value;
    }
    
    public final long accumulateAndGet(long n, LongBinaryOperator longBinaryOperator) {
        long value;
        long applyAsLong;
        do {
            value = this.get();
            applyAsLong = longBinaryOperator.applyAsLong(value, n);
        } while (!this.compareAndSet(value, applyAsLong));
        return applyAsLong;
    }
    
    @Override
    public String toString() {
        return Long.toString(this.get());
    }
    
    @Override
    public int intValue() {
        return (int)this.get();
    }
    
    @Override
    public long longValue() {
        return this.get();
    }
    
    @Override
    public float floatValue() {
        return (float)this.get();
    }
    
    @Override
    public double doubleValue() {
        return (double)this.get();
    }
    
    static {
        unsafe = Unsafe.getUnsafe();
        VM_SUPPORTS_LONG_CAS = VMSupportsCS8();
        try {
            valueOffset = AtomicLong.unsafe.objectFieldOffset(AtomicLong.class.getDeclaredField("value"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
}
